/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

package org.jboss.remoting.transport.socket;

import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.ServerInvoker;
import org.jboss.remoting.util.TimerUtil;
import org.jboss.remoting.marshal.serializable.SerializableMarshaller;
import org.jboss.util.propertyeditor.PropertyEditors;

import javax.net.ServerSocketFactory;
import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TimerTask;

/**
 * SocketServerInvoker is the server-side of a SOCKET based transport
 *
 * @author <a href="mailto:jhaynie@vocalocity.net">Jeff Haynie</a>
 * @author <a href="mailto:tom.elrod@jboss.com">Tom Elrod</a>
 * @version $Revision: 1.30.2.4 $
 * @jmx:mbean
 */
public class SocketServerInvoker extends ServerInvoker implements Runnable, SocketServerInvokerMBean
{
   private InetAddress addr;
   private int port;
   static int clientCount = 0;

   private Properties props = new Properties();

   private static int BACKLOG_DEFAULT = 200;
   protected static int MAX_POOL_SIZE_DEFAULT = 300;

   /**
    * Key for indicating if socket invoker should continue to keep socket connection between
    * client and server open after invocations by sending a ping on the connection
    * before being re-used.  The default for this is false.
    */
   public static final String CHECK_CONNECTION_KEY = "socket.check_connection";

   /**
    * Specifies the fully qualified class name for the custom SocketWrapper implementation to use on the server.
    */
   public static final String SERVER_SOCKET_CLASS_FLAG = "serverSocketClass";
   protected String serverSocketClass = ServerSocketWrapper.class.getName();

   protected ServerSocket serverSocket = null;
   protected boolean running = false;
   protected int backlog = BACKLOG_DEFAULT;
   protected Thread[] acceptThreads;
   protected int numAcceptThreads = 1;
   protected int maxPoolSize = MAX_POOL_SIZE_DEFAULT;
   protected LRUPool clientpool;
   protected LinkedList threadpool;

   protected boolean newServerSocketFactory = false;

   protected boolean reuseAddress = true;

   // defaults to -1 as to not have idle timeouts
   protected int idleTimeout = -1;
   protected IdleTimerTask idleTimerTask = null;

   /**
    * The logging trace level flag
    */
   protected boolean trace = false;


   public SocketServerInvoker(InvokerLocator locator)
   {
      super(locator);
   }

   public SocketServerInvoker(InvokerLocator locator, Map configuration)
   {
      super(locator, configuration);
   }

   /**
    * after a truststore update use this to
    * set a new ServerSocketFactory to the invoker<br>
    * then a new ServerSocket is created that accepts the new connections
    * @param serverSocketFactory
 	*/
   public void setNewServerSocketFactory(ServerSocketFactory serverSocketFactory){
	   newServerSocketFactory=true;
	   setServerSocketFactory(serverSocketFactory);
   }

   /**
    * refreshes the serverSocket by closing old one and
    * creating a new ServerSocket from new ServerSocketFactory
    * @throws IOException
    */
   protected void refreshServerSocket() throws IOException{
	   newServerSocketFactory=false;
	   serverSocket.close();
	   serverSocket=null;
	   InetAddress bindAddress = InetAddress.getByName(getServerBindAddress());
	   serverSocket = createServerSocket(getServerBindPort(), backlog, bindAddress);
      serverSocket.setReuseAddress(reuseAddress);
   }

   protected void setup()
         throws Exception
   {
      props.putAll(getConfiguration());
      PropertyEditors.mapJavaBeanProperties(this, props, false);

      super.setup();

      String ssclass = props.getProperty(SERVER_SOCKET_CLASS_FLAG);
      if(ssclass != null)
      {
         serverSocketClass = ssclass;
      }
   }

   protected void finalize() throws Throwable
   {
      stop();
      super.finalize();
   }

   /**
    * Starts the invoker.
    *
    * @jmx.managed-operation description = "Start sets up the ServerInvoker we are wrapping."
    * impact      = "ACTION"
    */
   public synchronized void start() throws IOException
   {

      trace = log.isTraceEnabled();

      if(!running)
      {
         InetAddress bindAddress = InetAddress.getByName(getServerBindAddress());

         if(maxPoolSize <= 0)
         {
            //need to reset to default
            maxPoolSize = MAX_POOL_SIZE_DEFAULT;
         }
         try
         {
            serverSocket = createServerSocket(getServerBindPort(), backlog, bindAddress);
            serverSocket.setReuseAddress(reuseAddress);
         }
         catch(IOException e)
         {
            log.error("Error starting ServerSocket.  Bind port: " + getServerBindPort() + ", bind address: " + bindAddress);
            throw e;
         }

         clientpool = new LRUPool(2, maxPoolSize);
         clientpool.create();
         threadpool = new LinkedList();

         acceptThreads = new Thread[numAcceptThreads];
         for(int i = 0; i < numAcceptThreads; i++)
         {
            if(trace)
            {
               log.trace("Creating another AcceptThread");
            }
            String name = getThreadName(i);
            acceptThreads[i] = new Thread(this, name);
            //acceptThreads[i].start();
            if(trace)
            {
               log.trace("Thread accepted");
            }
         }
      }
      try
      {
         super.start();
      }
      catch(IOException e)
      {
         log.error("Error starting SocketServerInvoker.", e);
         cleanup();
      }
      if(!running)
      {
         running = true;

         for(int i = 0; i < numAcceptThreads; i++)
         {
            acceptThreads[i].start();
         }
      }

      if(idleTimeout > 0)
      {
         if(idleTimerTask != null)
         {
            idleTimerTask.cancel();
         }
         idleTimerTask = new IdleTimerTask();
         TimerUtil.schedule(idleTimerTask, idleTimeout * 1000);
      }
      else
      {
         if(idleTimerTask != null)
         {
            idleTimerTask.cancel();
         }
      }

   }

   protected ServerSocket createServerSocket(int serverBindPort, int backlog, InetAddress bindAddress) throws IOException
   {
      return getServerSocketFactory().createServerSocket(serverBindPort, backlog, bindAddress);
   }

   protected String getThreadName(int i)
   {
      return "SocketServerInvoker#" + i + "-" + getServerBindPort();
   }

   public void destroy()
   {
      if(clientpool != null)
      {
         clientpool.destroy();
      }
      super.destroy();
   }

   /**
    * Stops the invoker.
    *
    * @jmx.managed-operation description = "Stops the invoker."
    * impact      = "ACTION"
    */
   public synchronized void stop()
   {

      if(running)
      {
         cleanup();
      }
      super.stop();
   }

   protected void cleanup()
   {
      running = false;

      maxPoolSize = 0; // so ServerThreads don't reinsert themselves
      if(acceptThreads != null)
      {
         for(int i = 0; i < acceptThreads.length; i++)
         {
            try
            {
               acceptThreads[i].interrupt();
            }
            catch(Exception ignored)
            {
            }
         }
      }

/*
 * The following code has been changed to avoid a race condition with ServerThread.run() which
 * can result in leaving ServerThreads alive, which causes a memory leak.
 */
      if (clientpool != null)
      {
         synchronized (clientpool)
         {
            Set svrThreads = clientpool.getContents();
            Iterator itr = svrThreads.iterator();

            while(itr.hasNext())
            {
               Object o = itr.next();
               ServerThread st = (ServerThread) o;
               st.shutdown();
            }

            clientpool.flush();
            clientpool.stop();

            if (threadpool != null)
            {
               synchronized(threadpool)
               {
                  int threadsToShutdown = threadpool.size();
                  for(int i = 0; i < threadsToShutdown; i++)
                  {
                     ServerThread thread = (ServerThread) threadpool.removeFirst();
                     thread.shutdown();
                  }
               }
            }
         }
      }

      try
      {
         serverSocket.close();
      }
      catch(Exception e)
      {
      }
   }

   /**
    * Indicates if SO_REUSEADDR is enabled on server sockets
    * Default is true.
    * @return
    */
   public boolean getReuseAddress()
   {
      return reuseAddress;
   }

   /**
    * Sets if SO_REUSEADDR is enabled on server sockets.
    * Default is true.
    *
    * @param reuse
    */
   public void setReuseAddress(boolean reuse)
   {
      this.reuseAddress = reuse;
   }

   /**
    * @return Value of property serverBindPort.
    * @jmx:managed-attribute
    */
   public int getCurrentThreadPoolSize()
   {
      return threadpool.size();
   }

   /**
    * @return Value of property serverBindPort.
    * @jmx:managed-attribute
    */
   public int getCurrentClientPoolSize()
   {
      return clientpool.size();
   }

   /**
    * Getter for property numAcceptThreads
    *
    * @return The number of threads that exist for accepting client connections
    * @jmx:managed-attribute
    */
   public int getNumAcceptThreads()
   {
      return numAcceptThreads;
   }

   /**
    * Setter for property numAcceptThreads
    *
    * @param size The number of threads that exist for accepting client connections
    * @jmx:managed-attribute
    */
   public void setNumAcceptThreads(int size)
   {
      this.numAcceptThreads = size;
   }

   /**
    * Setter for max pool size.
    * The number of server threads for processing client. The default is 300.
    *
    * @return
    * @jmx:managed-attribute
    */
   public int getMaxPoolSize()
   {
      return maxPoolSize;
   }

   /**
    * The number of server threads for processing client. The default is 300.
    *
    * @param maxPoolSize
    * @jmx:managed-attribute
    */
   public void setMaxPoolSize(int maxPoolSize)
   {
      this.maxPoolSize = maxPoolSize;
   }

   /**
    * @jmx:managed-attribute
    */
   public int getBacklog()
   {
      return backlog;
   }

   /**
    * @jmx:managed-attribute
    */
   public void setBacklog(int backlog)
   {
      if(backlog < 0)
      {
         this.backlog = BACKLOG_DEFAULT;
      }
      else
      {
         this.backlog = backlog;
      }
   }


   public int getIdleTimeout()
   {
      return idleTimeout;
   }

   /**
    * Sets the timeout for idle threads to be removed from pool.
    * If the value is greater than 0, then idle timeout will be
    * activated, otherwise no idle timeouts will occur.  By default,
    * this value is -1.
    *
    * @param idleTimeout number of seconds before a idle thread is timed out.
    */
   public void setIdleTimeout(int idleTimeout)
   {
      this.idleTimeout = idleTimeout;

      if(isStarted())
      {
         if(idleTimeout > 0)
         {
            if(idleTimerTask != null)
            {
               idleTimerTask.cancel();
            }
            idleTimerTask = new IdleTimerTask();
            TimerUtil.schedule(idleTimerTask, idleTimeout * 1000);
         }
         else
         {
            if(idleTimerTask != null)
            {
               idleTimerTask.cancel();
            }
         }
      }
   }

   public void run()
   {
      if(trace)
      {
         log.trace("Started execution of method run");
      }

      ServerSocketRefresh thread=new ServerSocketRefresh();
      thread.start();

      try
      {
         while(running)
         {
            try
            {
               if(trace)
               {
                  log.trace("Socket is going to be accepted");
               }

               thread.release(); //goes on if serversocket refresh is completed
               Socket socket = serverSocket.accept();
               if(trace)
               {
                  log.trace("Accepted: " + socket);
               }
               processInvocation(socket);
            }
            catch (javax.net.ssl.SSLException e)
            {
               log.error("SSLServerSocket error", e);
               return;
            }
            catch(Throwable ex)
            {
               if(running)
               {
                  log.error("Failed to accept socket connection", ex);
               }
            }
         }
      }
      finally
      {
         thread.interrupt();
      }
   }

   protected void processInvocation(Socket socket)
         throws Exception
   {
      ServerThread thread = null;
      boolean newThread = false;

      while(thread == null)
      {
         if(log.isTraceEnabled())
         {
            log.trace("try to get a thread for processing");
         }
         synchronized(threadpool)
         {
            if(threadpool.size() > 0)
            {
               thread = (ServerThread) threadpool.removeFirst();
               if(log.isTraceEnabled())
               {
                  log.trace("Got thread for processing - " + thread);
               }
            }
         }
         if(thread == null)
         {
            synchronized(clientpool)
            {
               if(clientpool.size() < maxPoolSize)
               {
                  thread = new ServerThread(socket, this, clientpool, threadpool, getTimeout(), serverSocketClass);
                  newThread = true;
               }
               if(thread == null)
               {
                  clientpool.evict();
                  if(trace)
                  {
                     log.trace("Waiting for a thread...");
                  }
                  clientpool.wait();
                  if(trace)
                  {
                     log.trace("Notified of available thread");
                  }
               }
            }
         }
      }
      synchronized(clientpool)
      {
         clientpool.insert(thread, thread);
      }

      if(newThread)
      {
         if(trace)
         {
            log.trace("Created a new thread, t=" + thread);
         }
         thread.start();
      }
      else
      {
         if(trace)
         {
            log.trace("Reusing thread t=" + thread);
         }
         thread.wakeup(socket, getTimeout(), locator.getParameters());
      }
   }

   /**
    * returns true if the transport is bi-directional in nature, for example,
    * SOAP in unidirectional and SOCKETs are bi-directional (unless behind a firewall
    * for example).
    *
    * @return
    */
   public boolean isTransportBiDirectional()
   {
      return true;
   }

   /**
    * Each implementation of the remote client invoker should have
    * a default data type that is uses in the case it is not specified
    * in the invoker locator uri.
    *
    * @return
    */
   protected String getDefaultDataType()
   {
      return SerializableMarshaller.DATATYPE;
   }

   /**
    * this thread checks if a new ServerSocketFactory was set,<br>
    * if so initializes a serversocket refresh
    * @author Michael Voss
    *
    */
   public class ServerSocketRefresh extends Thread{
	   boolean serverSocketLocked=false;

       public ServerSocketRefresh()
       {
          super("ServerSocketRefresh");
       }
	   public void run(){
 		  while(true){
     		  if(isInterrupted())break;
     		  if(newServerSocketFactory){
     			  log.debug("got notice about new ServerSocketFactory");
     			  serverSocketLocked=true;
				  try {
					  log.debug("refreshing server socket");
					  refreshServerSocket();
				  } catch (IOException e) {
					  log.debug("could not refresh server socket");
					  log.debug("message is: "+e.getMessage());
				  }
				  log.debug("server socket refreshed");
				  serverSocketLocked=false;

     		  }
     		  try {
     			  Thread.sleep(10000);
     		  } catch (InterruptedException e) {
				  interrupt();
     		  }
 		  }
 	  }

	/**
	 * let the run() method resume when refresh is completed
	 */
	public void release(){
 		  while (serverSocketLocked){
 			  try{Thread.sleep(1000);}catch(InterruptedException ignored){}
 		  }
 	  }
   }

   /**
    * The IdleTimerTask is used to periodically check the server threads to
    * see if any have been idle for a specified amount of time, and if so,
    * release those threads and their connections and clear from the server
    * thread pool.
    */
   public class IdleTimerTask extends TimerTask
   {
      public void run()
      {
         Object[] svrThreadArray = null;

         synchronized(clientpool)
         {
            Set svrThreads = clientpool.getContents();
            svrThreadArray = svrThreads.toArray();
         }
         if(trace)
         {
            if(svrThreadArray != null)
               {
                  log.trace("Idle timer task fired.  Number of ServerThreads = " + svrThreadArray.length);
               }
         }

         // iterate through pooled server threads and evict idle ones
         if(svrThreadArray != null)
         {
            long currentTime = System.currentTimeMillis();

            for(int x = 0; x < svrThreadArray.length; x++)
            {
               ServerThread svrThread = (ServerThread)svrThreadArray[x];

               // check the idle time and evict
               long idleTime = currentTime - svrThread.getLastRequestTimestamp();

               if(trace)
               {
                  log.trace("Idle time for ServerThread (" + svrThread + ") is " + idleTime);
               }

               long idleTimeout = getIdleTimeout() * 1000;
               if(idleTime > idleTimeout)
               {
                  if(trace)
                  {
                     log.trace("Idle timeout reached for ServerThread (" + svrThread + ") and will be evicted.");
                  }
                  clientpool.remove(svrThread);
                  svrThread.shutdown();
                  svrThread.unblock();
               }
            }
         }

         // now check idle server threads in the thread pool
         svrThreadArray = null;
         synchronized(threadpool)
         {
            if(threadpool.size() > 0)
            {
               // now need to check the tread pool to remove threads
               svrThreadArray = threadpool.toArray();
            }
         }

         if(trace)
         {
            if(svrThreadArray != null)
            {
               log.trace("Number of ServerThread in thead pool = " + svrThreadArray.length);
            }
         }

         if(svrThreadArray != null)
         {
            long currentTime = System.currentTimeMillis();

            for(int x = 0; x < svrThreadArray.length; x++)
            {
               ServerThread svrThread = (ServerThread)svrThreadArray[x];
               long idleTime = currentTime - svrThread.getLastRequestTimestamp();

               if(trace)
               {
                  log.trace("Idle time for ServerThread (" + svrThread + ") is " + idleTime);
               }

               long idleTimeout = getIdleTimeout() * 1000;
               if(idleTime > idleTimeout)
               {
                  if(trace)
                  {
                     log.trace("Idle timeout reached for ServerThread (" + svrThread + ") and will be removed from thread pool.");
                  }
                  threadpool.remove(svrThread);
                  svrThread.shutdown();
               }
            }
         }
      }
   }

}
